from unittest.mock import Mock, call

from baby_steps import given, then, when
from district42 import schema

from .._fixtures import *  # noqa: F401, F403
from .._utils import schema_mock


def test_any_generation(*, generate, random_):
    with given:
        sch = schema.any

    with when:
        res = generate(sch)

    with then:
        assert res is None
        assert random_.mock_calls == []


def test_any_type_generation(*, generate, generator, random_):
    with given:
        val = 42
        type_ = schema_mock(return_value=val)
        sch = schema.any(type_)

    with when:
        res = generate(sch)

    with then:
        assert res == val
        assert random_.mock_calls == [
            call.random_choice((type_,))
        ]
        assert type_.mock_calls == [call.__accept__(generator)]


def test_any_types_generation(*, generate, generator, random_):
    with given:
        val1, val2 = 42, "banana"
        type1_ = schema_mock(return_value=val1)
        type2_ = schema_mock(return_value=val2)
        sch = schema.any(type1_, type2_)
        random_.random_choice = Mock(return_value=type1_)

    with when:
        res = generate(sch)

    with then:
        assert res == val1
        assert random_.mock_calls == [
            call.random_choice((type1_, type2_))
        ]
        assert type1_.mock_calls == [call.__accept__(generator)]



from unittest.mock import call

from baby_steps import given, then, when
from district42 import optional, schema

from .._fixtures import *  # noqa: F401, F403
from .._utils import schema_mock


def test_dict_generation(*, generate, random_):
    with given:
        sch = schema.dict

    with when:
        res = generate(sch)

    with then:
        assert res == {}


def test_dict_empty_keys_generation(*, generate, random_):
    with given:
        sch = schema.dict({})

    with when:
        res = generate(sch)

    with then:
        assert res == {}


def test_dict_keys_generation(*, generate, generator, random_):
    with given:
        type1_ = schema_mock(return_value=42)
        type2_ = schema_mock(return_value="banana")
        sch = schema.dict({
            "id": type1_,
            "name": type2_,
        })

    with when:
        res = generate(sch)

    with then:
        assert res == {
            "id": 42,
            "name": "banana",
        }
        assert random_.mock_calls == []
        assert type1_.mock_calls == [call.__accept__(generator)]
        assert type2_.mock_calls == [call.__accept__(generator)]


def test_dict_relaxed_empty_keys_generation(*, generate, random_):
    with given:
        sch = schema.dict({...: ...})

    with when:
        res = generate(sch)

    with then:
        assert res == {}


def test_dict_relaxed_keys_generation(*, generate, generator, random_):
    with given:
        type1_ = schema_mock(return_value=42)
        type2_ = schema_mock(return_value="banana")
        sch = schema.dict({
            "id": type1_,
            "name": type2_,
            ...: ...
        })

    with when:
        res = generate(sch)

    with then:
        assert res == {
            "id": 42,
            "name": "banana",
        }
        assert random_.mock_calls == []
        assert type1_.mock_calls == [call.__accept__(generator)]
        assert type2_.mock_calls == [call.__accept__(generator)]


def test_dict_with_optional_key_generation(*, generate, generator, random_):
    with given:
        type1_ = schema_mock(return_value=42)
        type2_ = schema_mock(return_value="banana")
        sch = schema.dict({
            "id": type1_,
            optional("name"): type2_,
        })

    with when:
        res = generate(sch)

    with then:
        assert res == {"id": 42}
        assert random_.mock_calls == []
        assert type1_.mock_calls == [call.__accept__(generator)]
        assert type2_.mock_calls == []



from unittest.mock import call

from baby_steps import given, then, when
from district42 import schema

from blahblah._consts import INT_MAX, INT_MIN

from .._fixtures import *  # noqa: F401, F403


def test_int_generation(*, generate, random_):
    with given:
        sch = schema.int

    with when:
        res = generate(sch)

    with then:
        assert isinstance(res, int)
        assert INT_MIN <= res <= INT_MAX
        assert random_.mock_calls == [
            call.random_int(INT_MIN, INT_MAX)
        ]


def test_int_value_generation(*, generate, random_):
    with given:
        val = 42
        sch = schema.int(val)

    with when:
        res = generate(sch)

    with then:
        assert res == val
        assert random_.mock_calls == []


def test_int_min_generation(*, generate, random_):
    with given:
        min_val = 1
        sch = schema.int.min(min_val)

    with when:
        res = generate(sch)

    with then:
        assert res >= min_val
        assert random_.mock_calls == [
            call.random_int(min_val, INT_MAX)
        ]


def test_int_max_generation(*, generate, random_):
    with given:
        max_val = 2
        sch = schema.int.max(max_val)

    with when:
        res = generate(sch)

    with then:
        assert res <= max_val
        assert random_.mock_calls == [
            call.random_int(INT_MIN, max_val)
        ]


def test_int_min_max_generation(*, generate, random_):
    with given:
        min_val, max_val = 1, 2
        sch = schema.int.min(min_val).max(max_val)

    with when:
        res = generate(sch)

    with then:
        assert min_val <= res <= max_val
        assert random_.mock_calls == [
            call.random_int(min_val, max_val)
        ]



from typing import Any

from district42 import GenericSchema
from district42.types import Schema

from ._generator import Generator
from ._random import Random
from ._version import version

__version__ = version
__all__ = ("fake", "Generator", "Random",)

_generator = Generator(Random())


def fake(schema: GenericSchema, **kwargs: Any) -> Any:
    return schema.__accept__(_generator, **kwargs)


Schema.__override__(Schema.__invert__.__name__, fake)



import string
from typing import Any, Dict, List

from district42 import SchemaVisitor
from district42.types import (
    AnySchema,
    BoolSchema,
    ConstSchema,
    DictSchema,
    FloatSchema,
    IntSchema,
    ListSchema,
    NoneSchema,
    StrSchema,
)
from district42.utils import is_ellipsis
from niltype import Nil

from ._consts import (
    FLOAT_MAX,
    FLOAT_MIN,
    INT_MAX,
    INT_MIN,
    LIST_LEN_MAX,
    LIST_LEN_MIN,
    STR_LEN_MAX,
    STR_LEN_MIN,
)
from ._random import Random

__all__ = ("Generator",)


class Generator(SchemaVisitor[Any]):
    def __init__(self, random: Random) -> None:
        self._random = random

    def visit_none(self, schema: NoneSchema, **kwargs: Any) -> None:
        return None

    def visit_bool(self, schema: BoolSchema, **kwargs: Any) -> bool:
        if schema.props.value is not Nil:
            return schema.props.value
        return self._random.random_choice((True, False))

    def visit_int(self, schema: IntSchema, **kwargs: Any) -> int:
        if schema.props.value is not Nil:
            return schema.props.value

        min_value = schema.props.min if (schema.props.min is not Nil) else INT_MIN
        max_value = schema.props.max if (schema.props.max is not Nil) else INT_MAX
        return self._random.random_int(min_value, max_value)

    def visit_float(self, schema: FloatSchema, **kwargs: Any) -> float:
        if schema.props.value is not Nil:
            return schema.props.value

        min_value = schema.props.min if (schema.props.min is not Nil) else FLOAT_MIN
        max_value = schema.props.max if (schema.props.max is not Nil) else FLOAT_MAX
        return self._random.random_float(min_value, max_value)

    def visit_str(self, schema: StrSchema, **kwargs: Any) -> str:
        if schema.props.value is not Nil:
            return schema.props.value

        if schema.props.len is not Nil:
            length = schema.props.len
        else:
            min_length = schema.props.min_len if (schema.props.min_len is not Nil) else STR_LEN_MIN
            max_length = schema.props.max_len if (schema.props.max_len is not Nil) else STR_LEN_MAX
            length = self._random.random_int(min_length, max_length)

        if schema.props.alphabet is not Nil:
            alphabet = schema.props.alphabet
        else:
            alphabet = string.printable

        return self._random.random_str(length, alphabet)

    def visit_list(self, schema: ListSchema, **kwargs: Any) -> List[Any]:
        if schema.props.elements is not Nil:
            elements = []
            for elem in schema.props.elements:
                if is_ellipsis(elem):
                    continue
                elements.append(elem.__accept__(self, **kwargs))
            return elements

        is_length_specified = False
        if schema.props.len is not Nil:
            length = schema.props.len
            is_length_specified = True
        else:
            min_length = LIST_LEN_MIN
            if schema.props.min_len is not Nil:
                min_length = schema.props.min_len
                is_length_specified = True
            max_length = LIST_LEN_MAX
            if schema.props.max_len is not Nil:
                max_length = schema.props.max_len
                is_length_specified = True
            length = self._random.random_int(min_length, max_length)

        if schema.props.type is not Nil:
            return [schema.props.type.__accept__(self, **kwargs) for _ in range(length)]

        if is_length_specified:
            return [[] for _ in range(length)]
        return []

    def visit_dict(self, schema: DictSchema, **kwargs: Any) -> Dict[Any, Any]:
        generated: Dict[Any, Any] = {}
        if schema.props.keys is Nil:
            return generated

        for key, (val, is_optional) in schema.props.keys.items():
            if is_ellipsis(key):
                continue
            if is_optional:
                continue
            generated[key] = val.__accept__(self, **kwargs)

        return generated

    def visit_any(self, schema: AnySchema, **kwargs: Any) -> Any:
        if schema.props.types is Nil:
            return None
        chosen = self._random.random_choice(schema.props.types)
        return chosen.__accept__(self, **kwargs)

    def visit_const(self, schema: ConstSchema, **kwargs: Any) -> Any:
        if schema.props.value is Nil:
            return None
        return schema.props.value



from baby_steps import given, then, when
from district42 import schema

from .._fixtures import *  # noqa: F401, F403


def test_const_generation(*, generate, random_):
    with given:
        sch = schema.const

    with when:
        res = generate(sch)

    with then:
        assert res is None


def test_const_value_generation(*, generate, random_):
    with given:
        val = "banana"
        sch = schema.const(val)

    with when:
        res = generate(sch)

    with then:
        assert res == val



from unittest.mock import call

from baby_steps import given, then, when
from district42 import schema

from .._fixtures import *  # noqa: F401, F403


def test_bool_generation(*, generate, random_):
    with given:
        sch = schema.bool

    with when:
        res = generate(sch)

    with then:
        assert isinstance(res, bool)
        assert random_.mock_calls == [
            call.random_choice((True, False))
        ]


def test_bool_value_generation(*, generate, random_):
    with given:
        sch = schema.bool(True)

    with when:
        res = generate(sch)

    with then:
        assert res is True
        assert random_.mock_calls == []



import random
from typing import Any, List, Sequence, TypeVar

__all__ = ("Random",)

_T = TypeVar("_T")
SeedType = TypeVar("SeedType", int, float, str, bytes, bytearray)


class Random:
    def set_seed(self, seed: SeedType) -> None:
        random.seed(seed)

    def random_int(self, start: int, end: int) -> int:
        return random.randint(start, end)

    def random_float(self, start: float, end: float) -> float:
        return random.uniform(start, end)

    def random_str(self, length: int, alphabet: str) -> str:
        return "".join(random.choice(alphabet) for _ in range(length))

    def random_choice(self, sequence: Sequence[_T]) -> _T:
        return random.choice(sequence)

    def shuffle_list(self, elements: List[Any]) -> None:
        random.shuffle(elements)



from unittest.mock import call

from baby_steps import given, then, when
from district42 import schema

from blahblah._consts import LIST_LEN_MAX, LIST_LEN_MIN

from .._fixtures import *  # noqa: F401, F403
from .._utils import schema_mock


def test_list_generation(*, generate, random_):
    with given:
        sch = schema.list

    with when:
        res = generate(sch)

    with then:
        assert res == []
        assert random_.mock_calls == [
            # this value is not actually used (implementation specific)
            call.random_int(LIST_LEN_MIN, LIST_LEN_MAX)
        ]


def test_list_no_elements_generation(*, generate, generator, random_):
    with given:
        sch = schema.list([])

    with when:
        res = generate(sch)

    with then:
        assert res == []
        assert random_.mock_calls == []


def test_list_elements_generation(*, generate, generator, random_):
    with given:
        val1, val2 = 1, 2
        elem1_ = schema_mock(return_value=val1)
        elem2_ = schema_mock(return_value=val2)
        sch = schema.list([elem1_, elem2_])

    with when:
        res = generate(sch)

    with then:
        assert res == [val1, val2]
        assert random_.mock_calls == []
        assert elem1_.mock_calls == [call.__accept__(generator)]
        assert elem2_.mock_calls == [call.__accept__(generator)]


def test_list_contains_no_elements_generation(*, generate, generator, random_):
    with given:
        sch = schema.list([...])

    with when:
        res = generate(sch)

    with then:
        assert res == []
        assert random_.mock_calls == []


def test_list_contains_head_elements_generation(*, generate, generator, random_):
    with given:
        val1, val2 = 1, 2
        elem1_ = schema_mock(return_value=val1)
        elem2_ = schema_mock(return_value=val2)
        sch = schema.list([elem1_, elem2_, ...])

    with when:
        res = generate(sch)

    with then:
        assert res == [val1, val2]
        assert random_.mock_calls == []
        assert elem1_.mock_calls == [call.__accept__(generator)]
        assert elem2_.mock_calls == [call.__accept__(generator)]


def test_list_contains_tail_elements_generation(*, generate, generator, random_):
    with given:
        val1, val2 = 1, 2
        elem1_ = schema_mock(return_value=val1)
        elem2_ = schema_mock(return_value=val2)
        sch = schema.list([..., elem1_, elem2_])

    with when:
        res = generate(sch)

    with then:
        assert res == [val1, val2]
        assert random_.mock_calls == []
        assert elem1_.mock_calls == [call.__accept__(generator)]
        assert elem2_.mock_calls == [call.__accept__(generator)]


def test_list_contains_elements_generation(*, generate, generator, random_):
    with given:
        val1, val2 = 1, 2
        elem1_ = schema_mock(return_value=val1)
        elem2_ = schema_mock(return_value=val2)
        sch = schema.list([..., elem1_, elem2_, ...])

    with when:
        res = generate(sch)

    with then:
        assert res == [val1, val2]
        assert random_.mock_calls == []
        assert elem1_.mock_calls == [call.__accept__(generator)]
        assert elem2_.mock_calls == [call.__accept__(generator)]


def test_list_len_generation(*, generate, random_):
    with given:
        list_len = 10
        sch = schema.list.len(list_len)

    with when:
        res = generate(sch)

    with then:
        assert res == [[]] * list_len
        assert random_.mock_calls == []


def test_list_min_len_generation(*, generate, random_):
    with given:
        list_min_len = 1
        sch = schema.list.len(list_min_len, ...)

        list_len = 5
        random_.random_int.return_value = list_len

    with when:
        res = generate(sch)

    with then:
        assert res == [[]] * list_len
        assert random_.mock_calls == [call.random_int(list_min_len, LIST_LEN_MAX)]


def test_list_max_len_generation(*, generate, random_):
    with given:
        list_max_len = 10
        sch = schema.list.len(..., list_max_len)

        list_len = 5
        random_.random_int.return_value = list_len

    with when:
        res = generate(sch)

    with then:
        assert res == [[]] * list_len
        assert random_.mock_calls == [call.random_int(LIST_LEN_MIN, list_max_len)]


def test_list_min_max_len_generation(*, generate, random_):
    with given:
        list_min_len, list_max_len = 1, 10
        sch = schema.list.len(list_min_len, list_max_len)

        list_len = 5
        random_.random_int.return_value = list_len

    with when:
        res = generate(sch)

    with then:
        assert res == [[]] * list_len
        assert random_.mock_calls == [call.random_int(list_min_len, list_max_len)]


def test_list_of_elements_generation(*, generate, generator, random_):
    with given:
        val = 42
        type_ = schema_mock(return_value=val)
        sch = schema.list(type_)

        list_len = 5
        random_.random_int.return_value = list_len

    with when:
        res = generate(sch)

    with then:
        assert res == [val] * list_len
        assert random_.mock_calls == [call.random_int(LIST_LEN_MIN, LIST_LEN_MAX)]
        assert type_.mock_calls == [call.__accept__(generator)] * list_len


def test_list_of_elements_len_generation(*, generate, generator, random_):
    with given:
        val = 42
        type_ = schema_mock(return_value=val)
        list_len = 10
        sch = schema.list(type_).len(list_len)

    with when:
        res = generate(sch)

    with then:
        assert res == [val] * list_len
        assert random_.mock_calls == []
        assert type_.mock_calls == [call.__accept__(generator)] * list_len


def test_list_of_elements_min_max_len_generation(*, generate, generator, random_):
    with given:
        val = 42
        type_ = schema_mock(return_value=val)
        list_min_len, list_max_len = 1, 10
        sch = schema.list(type_).len(list_min_len, list_max_len)

        list_len = 5
        random_.random_int.return_value = list_len

    with when:
        res = generate(sch)

    with then:
        assert res == [val] * list_len
        assert random_.mock_calls == [call.random_int(list_min_len, list_max_len)]
        assert type_.mock_calls == [call.__accept__(generator)] * list_len



from unittest.mock import call

from baby_steps import given, then, when
from district42 import schema

from blahblah._consts import FLOAT_MAX, FLOAT_MIN

from .._fixtures import *  # noqa: F401, F403


def _test_float_generation(*, generate, random_):
    with given:
        sch = schema.float

    with when:
        res = generate(sch)

    with then:
        assert isinstance(res, float)
        assert FLOAT_MIN <= res <= FLOAT_MAX
        assert random_.mock_calls == [
            call.random_int(FLOAT_MIN, FLOAT_MAX)
        ]


def test_float_value_generation(*, generate, random_):
    with given:
        val = 3.14
        sch = schema.float(val)

    with when:
        res = generate(sch)

    with then:
        assert res == val
        assert random_.mock_calls == []


def _test_float_min_generation(*, generate, random_):
    with given:
        min_val = 3.14
        sch = schema.float.min(min_val)

    with when:
        res = generate(sch)

    with then:
        assert res >= min_val
        assert random_.mock_calls == [
            call.random_int(min_val, FLOAT_MAX)
        ]


def _test_float_max_generation(*, generate, random_):
    with given:
        max_val = 6.28
        sch = schema.float.max(max_val)

    with when:
        res = generate(sch)

    with then:
        assert res <= max_val
        assert random_.mock_calls == [
            call.random_int(FLOAT_MIN, max_val)
        ]


def test_float_min_max_generation(*, generate, random_):
    with given:
        min_val, max_val = 3.14, 6.28
        sch = schema.float.min(min_val).max(max_val)

    with when:
        res = generate(sch)

    with then:
        assert min_val <= res <= max_val
        assert random_.mock_calls == [
            call.random_float(min_val, max_val)
        ]



import string
from unittest.mock import call

from baby_steps import given, then, when
from district42 import schema

from blahblah._consts import STR_LEN_MAX, STR_LEN_MIN

from .._fixtures import *  # noqa: F401, F403


def test_str_generation(*, generate, random_):
    with given:
        sch = schema.str

        str_len = 25
        random_.random_int.return_value = str_len

    with when:
        res = generate(sch)

    with then:
        assert isinstance(res, str)
        assert STR_LEN_MIN <= len(res) <= STR_LEN_MAX
        assert random_.mock_calls == [
            call.random_int(STR_LEN_MIN, STR_LEN_MAX),
            call.random_str(str_len, string.printable)
        ]


def test_str_value_generation(*, generate, random_):
    with given:
        val = "banana"
        sch = schema.str(val)

    with when:
        res = generate(sch)

    with then:
        assert res == val
        assert random_.mock_calls == []


def test_str_len_generation(*, generate, random_):
    with given:
        str_len = 10
        sch = schema.str.len(str_len)

    with when:
        res = generate(sch)

    with then:
        assert len(res) == str_len
        assert random_.mock_calls == [
            call.random_str(str_len, string.printable)
        ]


def test_str_min_len_generation(*, generate, random_):
    with given:
        str_min_len = 1
        sch = schema.str.len(str_min_len, ...)

        str_len = 5
        random_.random_int.return_value = str_len

    with when:
        res = generate(sch)

    with then:
        assert str_min_len <= len(res) <= STR_LEN_MAX
        assert random_.mock_calls == [
            call.random_int(str_min_len, STR_LEN_MAX),
            call.random_str(str_len, string.printable)
        ]


def test_str_len_max_generation(*, generate, random_):
    with given:
        str_max_len = 10
        str_len = 5
        random_.random_int.return_value = str_len
        sch = schema.str.len(..., str_max_len)

    with when:
        res = generate(sch)

    with then:
        assert STR_LEN_MIN <= len(res) <= str_max_len
        assert random_.mock_calls == [
            call.random_int(STR_LEN_MIN, str_max_len),
            call.random_str(str_len, string.printable)
        ]


def test_str_len_min_max_generation(*, generate, random_):
    with given:
        str_min_len, str_max_len = 1, 10
        sch = schema.str.len(str_min_len, str_max_len)

        str_len = 5
        random_.random_int.return_value = str_len

    with when:
        res = generate(sch)

    with then:
        assert str_min_len <= len(res) <= str_max_len
        assert random_.mock_calls == [
            call.random_int(str_min_len, str_max_len),
            call.random_str(str_len, string.printable)
        ]


def test_str_alphabet_generation(*, generate, random_):
    with given:
        str_alphabet = string.digits
        sch = schema.str.alphabet(str_alphabet)

        str_len = 5
        random_.random_int.return_value = str_len

    with when:
        res = generate(sch)

    with then:
        assert STR_LEN_MIN <= len(res) <= STR_LEN_MAX
        assert random_.mock_calls == [
            call.random_int(STR_LEN_MIN, STR_LEN_MAX),
            call.random_str(str_len, str_alphabet)
        ]


def test_str_alphabet_len_generation(*, generate, random_):
    with given:
        str_len = 5
        str_alphabet = string.digits
        sch = schema.str.alphabet(str_alphabet).len(str_len)

    with when:
        res = generate(sch)

    with then:
        assert len(res) == str_len
        assert random_.mock_calls == [
            call.random_str(str_len, str_alphabet)
        ]


def test_str_alphabet_min_max_len_generation(*, generate, random_):
    with given:
        str_min_len, str_max_len = 1, 10
        str_alphabet = string.digits
        sch = schema.str.alphabet(str_alphabet).len(str_min_len, str_max_len)

        str_len = 5
        random_.random_int.return_value = str_len

    with when:
        res = generate(sch)

    with then:
        assert str_min_len <= len(res) <= str_max_len
        assert random_.mock_calls == [
            call.random_int(str_min_len, str_max_len),
            call.random_str(str_len, str_alphabet)
        ]
